"""AESCipher.py

AES Cipher class for encrypting and decrypting string data.
"""

import base64
import hashlib

from Cryptodome.Cipher import AES
from Cryptodome.Random import get_random_bytes
from Cryptodome.Util.Padding import pad, unpad

from ..core.Logger import Logger


logger = Logger(__name__)


def catch_value_error_exception(func):
    """Catch ValueError exceptions."""

    def wrapper(*args, **kwargs):
        """Wrap around func and catch ValueError exceptions."""
        try:
            result = func(*args, **kwargs)
            return result
        except ValueError as e:
            logger.log.exception("AESCipher error: %s", e)
            return None

    return wrapper


class AESCipher:
    """Implement AES Cipher Block Chaining encryption and decryption."""

    def __init__(self, key: str) -> None:
        """Make a fixed sha256 bit length key."""
        self.key = hashlib.sha256(key.encode("utf-8")).digest()

    @catch_value_error_exception
    def encrypt(self, raw_data: str) -> bytes:
        """Encrypt raw data."""
        logger.log.info("AESCipher: Encrypting data")
        encoded_data = pad(raw_data.encode("utf-8"), AES.block_size)
        iv = get_random_bytes(AES.block_size)
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return base64.b64encode(iv + cipher.encrypt(encoded_data))

    @catch_value_error_exception
    def decrypt(self, encrypted_data: bytes) -> bytes:
        """Decrypt encoded data."""
        logger.log.info("AESCipher: decrypting data")
        decoded_data = base64.b64decode(encrypted_data)
        iv = decoded_data[: AES.block_size]
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return unpad(cipher.decrypt(decoded_data[AES.block_size :]), AES.block_size)



"""sync_operations.py

Network synchronization enums.
"""

from enum import Enum


sync_operations = Enum(
    value="OPERATION",
    names=[
        ("PUSH_REQUEST", 1),
        ("PULL_REQUEST", 2),
        ("REJECT", 3),
        ("ACCEPT", 4),
        ("NO_DATA", 5),
    ],
)



"""TodoDatabase.py

This module implements the to-do database.
"""

import configparser
import sys
import threading

from pathlib import Path

from PyQt6.QtCore import QObject

from ..core import settings
from ..core.Logger import Logger
from ..net import tcp_server_lib, tcp_client_lib


logger = Logger(__name__)


class TodoDatabase(QObject):
    """Maintains a database of to-do lists."""

    def __init__(self):
        """Create a working database."""
        super().__init__()
        logger.log.info("Building the to-do database")
        self.initialized = False
        self.server_up = False

        # dictionary of to-do lists, which are lists of dictionaries
        self.todo_lists = {}
        self.todo_total = 0
        self.list_count = 0

        # keep statistics on 'active' to-do list
        self.active_list = ""
        self.todo_count = 0

        # create an ini config parser
        self.config = configparser.ConfigParser()
        if not Path.exists(settings.ini_fn):
            self.write_default_config()
        self.parse_config()

        # buffer size for sending/receiving data
        self.buf_size = 4096

        # create a server, and start it in a new thread
        # the server will create a new thread for each connection established
        self.db_server = None
        if settings.options["run"]:
            self.start_server()

        # create a client
        self.db_client = tcp_client_lib.DatabaseClient()

    def write_default_config(self):
        """Write the default configuration for To-Do."""
        logger.log.info("Writing default configuration to %s", settings.ini_fn)
        self.config["database"] = {}
        self.config["database"]["active_list"] = ""
        self.config["database"]["sort_key"] = "priority"
        self.config["database"]["reverse_sort"] = "no"
        self.config["server"] = {}
        self.config["server"]["key"] = "BewareTheBlackGuardian"
        self.config["server"]["run"] = "yes"
        self.config["server"]["address"] = "0.0.0.0"
        self.config["server"]["port"] = "5364"
        self.config["server"]["pull"] = "yes"
        self.config["server"]["push"] = "yes"

        try:
            with open(settings.ini_fn, "w", encoding="utf-8") as f:
                self.config.write(f)
        except IOError as e:
            logger.log.exception("Unable to write default config: %s", e)
            sys.exit(1)

    def write_config(self):
        """Write the current configuration options to config file."""
        logger.log.info("Writing configuration file %s", settings.ini_fn)
        self.config["database"]["active_list"] = settings.options["active_list"]
        self.config["database"]["sort_key"] = settings.options["sort_key"]
        if settings.options["reverse_sort"]:
            self.config["database"]["reverse_sort"] = "yes"
        else:
            self.config["database"]["reverse_sort"] = "no"
        self.config["server"]["key"] = settings.options["key"]
        if settings.options["run"]:
            self.config["server"]["run"] = "yes"
        else:
            self.config["server"]["run"] = "no"
        self.config["server"]["address"] = settings.options["address"]
        self.config["server"]["port"] = str(settings.options["port"])
        self.config["server"]["pull"] = settings.options["pull"]
        self.config["server"]["push"] = settings.options["push"]

        try:
            with open(settings.ini_fn, "w", encoding="utf-8") as f:
                self.config.write(f)
                msg = f"Successfully wrote configuration file {settings.ini_fn}"
                logger.log.info(msg)
                return True, msg
        except IOError as e:
            msg = f"Unable to write default configuration file {settings.ini_fn}: {e}"
            logger.log.exception(msg)
            return False, msg

    def parse_config(self):
        """Parse configuration file.

        If a configuration file exists read it in and set relevant data.
        """
        logger.log.info("Parsing configuration file %s", settings.ini_fn)
        self.config.read(settings.ini_fn)

        for k, v in self.config["database"].items():
            if k not in settings.options:
                logger.log.info("%r = %r", k, v)
                settings.options[k] = v

        for k, v in self.config["server"].items():
            if k not in settings.options:
                logger.log.info("%r = %r", k, v)
                settings.options[k] = v

        # fix some option types
        if settings.options["reverse_sort"] == "yes":
            settings.options["reverse_sort"] = True
        elif settings.options["reverse_sort"] == "no":
            settings.options["reverse_sort"] = False
        else:
            logger.log.warning("Reverse sort option invalid, defaulting to no")
            settings.options["reverse_sort"] = False

        if settings.options["run"] == "yes":
            settings.options["run"] = True
        elif settings.options["run"] == "no":
            settings.options["run"] = False
        else:
            logger.log.warning("Run network server option invalid, defaulting to yes")
            settings.options["run"] = True

        try:
            settings.options["port"] = int(settings.options["port"])
        except ValueError as e:
            logger.log.exception("Port must be a number: %s", e)
            settings.options["port"] = 5364

    def write_text_file(self, fn=None):
        """Write active list to plain text file."""
        if self.todo_count == 0:
            return

        if fn is None:
            return

        logger.log.info("Writing pytodo-qt list to file %s", fn)
        try:
            with open(fn, "w", encoding="utf-8") as f:
                f.write(f"{self.active_list:*^60}\n\n")
                for todo in self.todo_lists[self.active_list]:
                    f.write(f'{todo["reminder"]}\n')
        except IOError as e:
            logger.log.exception("Unable to write pytodo-qt list to %s: %s", fn, e)

    def server_running(self):
        """Determine if the server is running"""
        if self.server_up and self.db_server is not None:
            return True

        return False

    def start_server(self):
        """Create and start the server."""
        if self.server_running():
            return

        logger.log.info("Starting the server")
        self.db_server = tcp_server_lib.DatabaseServer(
            (settings.options["address"], settings.options["port"]),
            tcp_server_lib.TCPRequestHandler,
        )

        # start network server thread
        st = threading.Thread(target=self.db_server.serve_forever)
        st.daemon = True
        st.start()
        self.server_up = True

        logger.log.info(
            "Server up at %s:%d", settings.options["address"], settings.options["port"]
        )

    def stop_server(self):
        """Stop and destroy the server."""
        if self.db_server is not None and self.server_running():
            logger.log.info("Shutting down the server")
            self.db_server.shutdown()
            self.db_server = None
            self.server_up = False
            logger.log.info("Server shut down")
        else:
            logger.log.exception(
                "network server does not exist, or is not running, exiting"
            )
            sys.exit(1)

    def restart_server(self):
        """Stop and then restart the server."""
        if self.server_running():
            self.stop_server()
            self.start_server()
        else:
            self.start_server()

    def sync_pull(self, host):
        """Perform a client pull."""
        return self.db_client.sync_pull(host)

    def sync_push(self, host):
        """Perform a client push."""
        return self.db_client.sync_push(host)

    def sort_active_list(self):
        """Sort the active to-do list."""
        logger.log.info("Sorting list %s", self.active_list)
        self.todo_lists[self.active_list].sort(
            key=lambda todo_list: todo_list[settings.options["sort_key"]],
            reverse=settings.options["reverse_sort"],
        )

    def todo_index(self, reminder):
        """Return the index location of the to-do matching reminder."""
        i = 0
        for todo in self.todo_lists[self.active_list]:
            if todo["reminder"] == reminder:
                return i
            i += 1

        return None



"""settings.py

This module creates To-Do core global variables and functions.
"""

import sys

from pathlib import Path

from ..core.Logger import Logger


logger = Logger(__name__)


__version__ = "0.2.8"
options = {}
DB = None

home_dir = Path.home()
if home_dir is not None:
    app_dir = Path.joinpath(home_dir, ".pytodo-qt")
else:
    logger.log.exception("Unable to get home directory, exiting")
    sys.exit(1)

if not Path.exists(app_dir):
    try:
        Path.mkdir(app_dir)
    except OSError as e:
        logger.log.exception("Error creating pytodo-qt configuration directory: %s", e)
        sys.exit(1)


# private files
ini_fn = Path.joinpath(app_dir, "pytodo-qt.ini")
db_fn = Path.joinpath(app_dir, "pytodo-qt-db.json")



"""SyncDialog.py

Simple dialog to collect information needed to perform a sync operation.
"""

from PyQt6.QtWidgets import (
    QDialog,
    QLabel,
    QLineEdit,
    QPushButton,
    QVBoxLayout,
    QMessageBox,
)
from PyQt6.QtGui import QIntValidator

from ..core import error_on_none_db, settings
from ..core.Logger import Logger
from ..net.sync_operations import sync_operations


logger = Logger(__name__)


class SyncDialog(QDialog):
    """Synchronize lists with remotes."""

    def __init__(self, operation=None):
        """Create a simple dialog.

        Get information about a remote host to perform a sync operation with.
        """
        logger.log.info("Creating a Sync dialog")

        super().__init__()

        self.operation = operation

        # address
        address_label = QLabel("Host Address", self)
        self.address_field = QLineEdit(self)

        # port
        port_label = QLabel("Port", self)
        self.port_field = QLineEdit(self)
        self.port_field.setText(str(settings.options["port"]))
        self.port_field.setValidator(QIntValidator())

        # add button
        self.get_button = QPushButton("Synchronize", self)
        self.get_button.clicked.connect(self.get_host)

        # create a vertical box layout
        v_box = QVBoxLayout()
        v_box.addWidget(address_label)
        v_box.addWidget(self.address_field)
        v_box.addWidget(port_label)
        v_box.addWidget(self.port_field)
        v_box.addWidget(self.get_button)

        # set layout and window title
        self.setLayout(v_box)
        self.setWindowTitle(f"Sync {operation}")
        self.setMinimumWidth(350)

        logger.log.info("Sync dialog created")

    @error_on_none_db
    def get_host(self, *args, **kwargs):
        """Get host information."""
        address = self.address_field.text()
        if address == "":
            QMessageBox.information(self, "Empty address", "Address cannot be empty")
            return

        # get remote port
        port = self.port_field.text()
        if port == "":
            QMessageBox.information(self, "Empty port", "Port cannot be empty")
            return

        port = int(port)

        logger.log.info("Got host information for sync operation")
        if self.operation == sync_operations["PULL_REQUEST"].name:
            # try the pull, inform user of the results
            result, msg = settings.DB.sync_pull((address, port))
            logger.log.info(msg)
            if not result:
                return
        elif self.operation == sync_operations["PUSH_REQUEST"].name:
            # temporarily enable pulling for the push
            pull_ok = settings.options["pull"]
            if not pull_ok:
                settings.options["pull"] = True

            # try the push, inform user of results
            logger.log.info("Sending a sync push request to %s:%d", address, port)
            result, msg = settings.DB.sync_push((address, port))
            logger.log.info(msg)
            if not pull_ok:
                settings.options["pull"] = False

            if not result:
                return

        self.accept()



"""tcp_client_lib.py

This module implements the To-Do database network client.
"""

import json
import socket
import time

from pathlib import Path

from PyQt6.QtCore import QObject, pyqtSignal

from ..core import settings, json_helpers
from ..core.Logger import Logger
from ..crypto.AESCipher import AESCipher
from ..net import recv_all
from ..net.sync_operations import sync_operations


logger = Logger(__name__)


class DatabaseClient(QObject):
    """To-Do database client class."""

    sync_occurred = pyqtSignal(str)

    def __init__(self):
        """Initialize client."""
        super().__init__()
        self.buf_size = 4096
        self.aes_cipher = AESCipher(settings.options["key"])

    def send_request(self, host, sock, request):
        """Send a request to remote connection."""
        logger.log.info("sending %s to %s", request, host)
        encrypted_request = self.aes_cipher.encrypt(request)
        sock.send(encrypted_request)

    def get_response(self, sock):
        """Get a response from remote connection."""
        encrypted_data = sock.recv(self.buf_size)
        decrypted_data = self.aes_cipher.decrypt(encrypted_data)
        response = decrypted_data.decode("utf-8")
        return response

    def process_response(self, host, sock, request, response):
        """Process remote host's response to a request."""
        msg = f"{host} responded to {request} with {response}"
        logger.log.info(msg)
        self.sync_occurred.emit(msg)

        if response == sync_operations["ACCEPT"].name:
            if request == sync_operations["PULL_REQUEST"].name:
                size_header = sock.recv(self.buf_size)
                decrypted_header = self.aes_cipher.decrypt(size_header)
                size = int(decrypted_header)
                logger.log.info("remote lists is %d bytes", size)
                time.sleep(1)
                data = recv_all(sock, size)
                self.process_data(host, data)
            elif request == sync_operations["PUSH_REQUEST"].name:
                pass
            else:
                return False, msg

        return True, msg

    def process_data(self, host, data):
        """Process encrypted data received."""
        # decrypt and decode received data
        decrypted_data = self.aes_cipher.decrypt(data)
        try:
            deserialized = json.loads(decrypted_data)
        except OSError as e:
            logger.log.exception(e)
            return False, e

        # write data to a temporary file, then read it in
        tmp = Path.joinpath(settings.app_dir, ".todo_lists.tmp")
        try:
            with open(tmp, "w", encoding="utf-8") as f:
                f.write(deserialized)
            msg = f"Pull from {host} successful."
            logger.log.info(msg)
            result, e = json_helpers.read_json_data(tmp)
            if not result:
                return False, e
            Path.unlink(tmp)
            self.sync_occurred.emit(msg)
            return True, msg
        except IOError as e:
            msg = f"Unable to write temporary file: {e}"
            logger.log.exception(msg)
            return False, msg

    def synchronize(self, host, request):
        """Synchronize to-do lists with other hosts."""
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                sock.settimeout(30)
                try:
                    sock.connect(host)
                except socket.error as err:
                    msg = f"Unable to connect to host: {err}"
                    logger.log.exception(msg)
                    return False, msg
                self.send_request(host, sock, request)
                response = self.get_response(sock)
                return self.process_response(host, sock, request, response)
        except OSError as e:
            msg = f"Unable to connect to host {host}: {e}"
            logger.log.exception(msg)
            return False, msg

    def sync_pull(self, host):
        """Synchronize database with another by pulling it from a host."""
        logger.log.info("Performing a Sync Pull")
        result, msg = self.synchronize(host, sync_operations["PULL_REQUEST"].name)
        self.sync_occurred.emit(f"PULL_REQUEST sent to {host}")
        return result, msg

    def sync_push(self, host):
        """Synchronize lists between devices by pushing them to a host.

        To-Do doesn't really support pushing because that is rude.
        In our terminology, if you do a push, it really means asking
        the device you want your to-do lists on to pull from you.
        """
        logger.log.info("Performing a Sync Push")
        result, msg = self.synchronize(host, sync_operations["PUSH_REQUEST"].name)
        self.sync_occurred.emit(f"PUSH_REQUEST sent to {host}")
        return result, msg



"""__main__.py

pytodo-qt

A to-do list program written in Python using PyQt6

Copyright (C) 2024 Michael Berry <trismegustis@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import argparse
import os
import sys

from PyQt6.QtWidgets import QApplication

from .core.Logger import Logger
from .core import settings
from .core.TodoDatabase import TodoDatabase
from .gui.MainWindow import MainWindow


logger = Logger(__name__)


# Main function
def main():
    # move to main module dir
    os.chdir(os.path.dirname(__file__))

    # create a command line arg_parser
    arg_parser = argparse.ArgumentParser(
        prog="pytodo-qt",
        description="To-Do List Application written in Python 3 and PyQt6",
        epilog="Copyright Michael Berry 2024",
    )

    # add network server command group
    server_group = arg_parser.add_argument_group("Server Commands")

    # add options
    server_group.add_argument(
        "-s",
        "--run-server",
        action="store",
        type=str,
        choices=["yes", "no"],
        help="run a simple to-do list network server",
    )

    server_group.add_argument(
        "-a",
        "--allow-pull",
        action="store",
        type=str,
        choices=["yes", "no"],
        help="allow remote users to copy your lists",
    )

    server_group.add_argument(
        "-A",
        "--allow-push",
        action="store",
        type=str,
        choices=["yes", "no"],
        help="allow remote users to send you their lists",
    )

    server_group.add_argument(
        "-i", "--ip", type=str, help="set the database server's IP address."
    )

    server_group.add_argument(
        "-p",
        "--port",
        type=int,
        help="specify which port the database server will bind to",
    )

    arg_parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s v{settings.__version__}"
    )

    # parse args then convert to dict format
    args = arg_parser.parse_args()
    for k, v in vars(args).items():
        if v is not None:
            settings.options[k] = v

    # create a QApplication, the main window, DB, then hand over control to Qt
    app = QApplication(sys.argv)
    settings.DB = TodoDatabase()
    _ = MainWindow()
    sys.exit(app.exec())



"""__init__.py

pytodo-qt.core: A decorator to make security checks on the to-do database.
"""

import sys

from PyQt6.QtWidgets import QMessageBox

from ..core import settings
from ..core.Logger import Logger


def error_on_none_db(func):
    """Check if settings.db is valid and run func or error out if it is None."""

    logger = Logger(func.__name__)

    def wrapper(*args, **kwargs):
        """Wrap around func and check settings.db"""

        if settings.DB is not None:
            try:
                result = func(*args, **kwargs)
                return result
            except OSError as e:
                logger.log.exception(f"To-Do error: {e}")
                return
        else:
            msg = "Database does not exist, exiting"
            QMessageBox.critical(None, "Database Error", msg)
            logger.log.exception(msg)
            sys.exit(1)

    return wrapper



"""AddTodoDialog.py

Simple dialog to create a to-do.
"""

from PyQt6.QtWidgets import (
    QDialog,
    QLabel,
    QLineEdit,
    QComboBox,
    QPushButton,
    QVBoxLayout,
    QMessageBox,
)

from ..core import error_on_none_db, settings
from ..core.Logger import Logger


logger = Logger(__name__)


class AddTodoDialog(QDialog):
    """Create a new to-do."""

    def __init__(self):
        """Create a simple dialog.

        Get enough information to create a new to-do.
        """
        logger.log.info("Creating an add to-do dialog")

        super().__init__()

        # reminder
        reminder_label = QLabel("Reminder", self)
        self.reminder_field = QLineEdit(self)

        # priority
        priority_label = QLabel("Priority", self)
        self.priority_field = QComboBox(self)
        self.priority_field.addItems(["Low", "Normal", "High"])

        # add button
        self.add_button = QPushButton("Add to-do", self)
        self.add_button.clicked.connect(self.get_todo)

        # create a vertical box layout
        v_box = QVBoxLayout()
        v_box.addWidget(reminder_label)
        v_box.addWidget(self.reminder_field)
        v_box.addWidget(priority_label)
        v_box.addWidget(self.priority_field)
        v_box.addWidget(self.add_button)

        # set layout and window title
        self.setLayout(v_box)
        self.setWindowTitle("Add New to-do")
        self.setMinimumWidth(500)

        logger.log.info("Add to-do dialog created")

    @error_on_none_db
    def get_todo(self, *args, **kwargs):
        """Get to-do information and append it to the current list."""
        reminder = self.reminder_field.text()
        if reminder == "":
            QMessageBox.information(self, "Empty reminder", "Reminder cannot be empty")
            return

        # get to-do information
        todo = {"complete": False, "reminder": reminder}
        priority = self.priority_field.currentText()
        if priority == "High":
            todo["priority"] = 1
        elif priority == "Normal":
            todo["priority"] = 2
        else:
            todo["priority"] = 3

        # update the database
        if settings.DB.todo_lists is not None and settings.DB.active_list is not None:
            settings.DB.todo_lists[settings.DB.active_list].append(todo)
            settings.DB.todo_count += 1
            settings.DB.todo_total += 1
        else:
            logger.log.exception(
                "Error: settings.db.todo_list or setting.db.active list does not exist, exiting"
            )

        logger.log.info("New to-do created: %s", todo)

        self.accept()



"""Logger.py

A Generic logging class.
"""

import logging
import sys

from pathlib import Path

home_dir = Path.home()
if home_dir is not None:
    app_dir = Path.joinpath(home_dir, ".pytodo-qt")
else:
    print("Error: unable to get home directory, exiting")
    sys.exit(1)

if not Path.exists(app_dir):
    try:
        Path.mkdir(app_dir)
    except OSError as e:
        print(f"Error: failed to create pytodo-qt directory {app_dir}: {e}")
        sys.exit(1)

log_fn = Path.joinpath(app_dir, "pytodo-qt.log")

if not Path.exists(log_fn):
    try:
        log_fn.touch()
    except OSError as e:
        print(f"Error: failed to create pytodo-qt log file {log_fn}: {e}")
        sys.exit(1)


class Logger:
    """A flexible logger."""

    def __init__(
        self,
        module_name,
        file_name=log_fn,
        file_mode="w",
        file_handler_level=logging.DEBUG,
        file_handler_format="%(asctime)-15s [%(threadName)-12s][%(levelname)-8s]  %(message)s",
        console_handler=logging.StreamHandler(),
        console_handler_level=logging.DEBUG,
        console_handler_format=logging.Formatter(
            "%(asctime)-15s [%(threadName)-12s][%(levelname)-8s]  %(message)s"
        ),
    ):
        """Initialize the logger."""
        logging.basicConfig(
            level=file_handler_level,
            format=file_handler_format,
            filename=file_name,
            filemode=file_mode,
        )

        self.log = logging.getLogger(module_name)
        console_handler.setLevel(console_handler_level)
        console_handler.setFormatter(console_handler_format)
        self.log.addHandler(console_handler)
